#
# Copyright 2018-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import division
import sys
from PIL import Image
import collections

from . import image_functions as F

if sys.version_info < (3, 3):
    Sequence = collections.Sequence
    Iterable = collections.Iterable
else:
    Sequence = collections.abc.Sequence
    Iterable = collections.abc.Iterable


class ImageProcessor(object):
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): sequence of transforms to compose.

    Example:
        >>> pipeline = ImageProcessor([
        >>>     Rotate(150),
        >>>     Resize([100,100])
        >>> ])
        >>> pipeline.apply_transforms(img)
    """

    def __init__(self, transforms=[]):
        assert isinstance(transforms, Sequence)  # nosec - assert
        self.transforms = transforms

    def apply_transforms(self, img):
        """
        Sequentially apply the list of transformations to the input image.

        args:
            img: an image in bytes format, as a Pillow image object, or a numpy ndarray

        output:
            The transformed image.
            Depending on the transformation the output is either a Pillow Image object or a numpy ndarray.
        """
        # verify whether the Normalize or Standardize transformations are positioned at the end
        encoding = [(isinstance(t, Normalize) or isinstance(t, Standardize)) for t in self.transforms]
        if sum(encoding[:-1]) != 0:
            raise ValueError('A Standardize or Normalize transformation can only be positioned at the end of the'
                             'pipeline.')
        # apply the transformations
        for t in self.transforms:
            img = t(img)
        return img


class ToPILImage(object):
    """Convert a byte stream or an ndarray to PIL Image.

    Converts a byte stream or a numpy ndarray of shape
    H x W x C to a PIL Image while preserving the value range.

    Args:
        mode (`PIL.Image mode`_): color space and pixel depth of input data (optional).
            If ``mode`` is ``None`` (default) there are some assumptions made about the input data:
             - If the input has 4 channels, the ``mode`` is assumed to be ``RGBA``.
             - If the input has 3 channels, the ``mode`` is assumed to be ``RGB``.
             - If the input has 2 channels, the ``mode`` is assumed to be ``LA``.
             - If the input has 1 channel, the ``mode`` is determined by the data type (i.e ``int``, ``float``,
              ``short``).

    .. _PIL.Image mode: https://pillow.readthedocs.io/en/latest/handbook/concepts.html#concept-modes
    """
    def __init__(self, target_mode, mode=None):
        self.mode = mode
        self.target_mode = target_mode

    def __call__(self, pic):
        """
        Args:
            pic (bytestream or numpy.ndarray): Image to be converted to PIL Image.

        Returns:
            PIL Image: Image converted to PIL Image.

        """
        return F.to_pil_image(pic, self.target_mode, self.mode)


class PILtoarray(object):
    """
    Convert a PIL Image object to a numpy ndarray.
    """

    def __call__(self, pic):
        """
        Args:
            pic (PIL Image): Image to be converted to a numpy ndarray.

        Returns:
            numpy ndarray
        """
        return F.pil_to_array(pic)


class Normalize(object):
    """
    Normalize the image to a range between [0, 1].
    """

    def __call__(self, img):
        """
        Args:
        img (PIL image or numpy.ndarray): Image to be normalized.

        Returns:
        numpy.ndarray: Normalized image.
        """
        return F.normalize(img)


class Standardize(object):
    """
    Standardize the image (mean-centering and STD of 1).

    Args:
        mean (optional): a single number or an n-dimensional sequence with n equal to the number of image channels
        std (optional): a single number or an n-dimensional sequence with n equal to the number of image channels
    Returns:
        numpy.ndarray: standardized image

    If `mean` or `std` are not provided, the channel-wise values will be calculated for the input image.
    """
    def __init__(self, mean=None, std=None):
        self.mean = mean
        self.std = std

    def __call__(self, img):
        """
        Args:
        img (PIL image or numpy.ndarray): Image to be standardized.

        Returns:
        numpy.ndarray: Standardized image.
        """
        return F.standardize(img, self.mean, self.std)


class Resize(object):
    """Resize the input PIL Image to the given size.

    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        assert isinstance(size, int) or (isinstance(size, Sequence) and len(size) == 2)  # nosec - assert
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be scaled.

        Returns:
            PIL Image: Rescaled image.
        """
        return F.resize(img, self.size, self.interpolation)


class Rotate(object):
    """
    Rotate the input PIL Image by a given angle (counter clockwise).

    Args:
        angle (int or float): Counter clockwise angle to rotate the image by.
    """

    def __init__(self, angle):
        self.angle = angle

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be rotated.

        Returns:
            PIL Image: Rotated image.
        """
        return F.rotate(img, self.angle)


class Grayscale(object):
    """Convert image to grayscale.

    Args:
        num_output_channels (int): (1, 3 or 4) number of channels desired for output image

    Returns:
        PIL Image: Grayscale version of the input.
        - If num_output_channels == 1 : returned image is single channel
        - If num_output_channels == 3 : returned image is 3 channel with r == g == b
        - If num_output_channels == 4 : returned image is 3 channel with r == g == b == a

    """

    def __init__(self, num_output_channels=1):
        self.num_output_channels = num_output_channels

    def __call__(self, img):
        """
        Args:
            img (PIL Image): Image to be converted to grayscale.

        Returns:
            PIL Image: Randomly grayscaled image.
        """
        return F.to_grayscale(img, num_output_channels=self.num_output_channels)



#
# Copyright 2018-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from flask import abort
import cyclonefw
from cyclonefw.utils.image_utils import ImageProcessor
import os, logging
from logging.handlers import RotatingFileHandler
import json
import requests


def redirect_errors_to_flask(func):
    """
    This decorator function will capture all Pythonic errors and return them as flask errors.

    If you are looking to disable this functionality, please remove this decorator from the `apply_transforms()` module
    under the ImageProcessor class.
    """

    def inner(*args, **kwargs):
        try:
            # run the function
            return func(*args, **kwargs)
        except ValueError as ve:
            if 'pic should be 2 or 3 dimensional' in str(ve):
                abort(400, "Invalid input, please ensure the input is either "
                           "a grayscale or a colour image.")
        except TypeError as te:
            if 'bytes or ndarray' in str(te):
                abort(400, "Invalid input format, please make sure the input file format "
                           " is a common image format such as JPG or PNG.")

    return inner


class MAXImageProcessor(ImageProcessor):
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> pipeline = ImageProcessor([
        >>>     Rotate(150),
        >>>     Resize([100,100])
        >>> ])
        >>> pipeline.apply_transforms(img)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @redirect_errors_to_flask
    def apply_transforms(self, img):
        return super().apply_transforms(img)


class AppFilter(logging.Filter):
    def __init__(self):
        pass

    def filter(self, record):
        record.app_name = cyclonefw.core.app.context.app_name
        try:
            trace_id = cyclonefw.core.app.getTraceId()
        except Exception as e:
            print(e)
        else:
            record.trace_id = trace_id

        return True


def getLogger(name="cyclonefw", formatter="[%(asctime)s] [%(pathname)s(line:%(lineno)d)] [%(levelname)s] %(message)s",
              logPath=None, level=logging.DEBUG):
    """
    获取logger

    日志记录使用如下格式: %s 等作为占位符

    print "My name is %s and weight is %d kg!" % ('Zara', 21)

    :param name: 日志名称
    :param formatter: 日志格式
    :param logPath: 日志目录, 默认为 logs目录下, 使用 . 可以指定当前执行目录
    :param level: 日志级别
    :return:
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if not logger.hasHandlers():
        consolelog = logging.StreamHandler()
        consolelog.setFormatter(logging.Formatter(formatter))
        logger.addHandler(consolelog)
        logger.addFilter(AppFilter())

    if logPath != None:
        logger.addHandler(logHandler(name, formatter, logPath))

    return logger


def logHandler(name, logPath="logs",
               formatter="[%(asctime)s] [%(pathname)s(line:%(lineno)d)] [%(levelname)s] %(message)s"):
    if not os.path.exists(logPath):
        os.makedirs(logPath, exist_ok=True)
    handler = CyclonefwRotatingFileHandler("{}/{}.log".format(logPath, name), "a", 1024 * 1024 * 20, 10)
    handler.setFormatter(logging.Formatter(formatter))
    return handler


class CyclonefwRotatingFileHandler(RotatingFileHandler):
    def emit(self, record):
        if hasattr(record, "trace_id"):
            self.setFormatter(logging.Formatter("[%(asctime)s] [%(trace_id)s] [%(app_name)s] [%(pathname)s(line:%(lineno)d)] [%(levelname)s] %(message)s"))

        super(CyclonefwRotatingFileHandler, self).emit(record)


def register(consulServer, name, ip, port, tags):
    """
    注册到注册中心
    :param consulServer:
    :param name:
    :param ip:
    :param port:
    :param tags:
    :return:
    """
    data = {
        "name": name,
        "address": ip,
        "port": port,
        "tags": tags,
        "check": {}
    }
    try:
        requests.put("{}/v1/agent/service/register".format(consulServer), headers={
            "Content-Type": "application/json"
        }, data=json.dumps(data))
    except Exception as e:
        print(e)


def unregister(consulServer, name):
    """
    反注册到注册中心
    :param consulServer:
    :param name:
    :return:
    """
    try:
        requests.put("{}/v1/agent/service/deregister/{}".format(consulServer, name))
    except Exception as e:
        print(e)


def getHostnameAndIp():
    """
    获取hostname及ip地址
    :return: hostname, ip
    """
    import socket
    hostname = socket.gethostname()
    ip = socket.gethostbyname(hostname)
    return hostname, ip



#
# Copyright 2018-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cyclonefw',
    version='1.0.17',
    description='A package to simplify the creation of MAX models',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/wenlaizhou/cyclonefw',
    author='CODAIT',
    author_email='',
    license='Apache',
    packages=['cyclonefw', 'cyclonefw.core', 'cyclonefw.model', 'cyclonefw.utils'],
    zip_safe=True,
    install_requires=[
        'requests',
        'flask==1.1.2',
        'werkzeug==0.16.1',
        'flask-restplus',
        'flask-restx==0.1.1',
        'flask-cors==3.0.9',
        'Pillow==7.2.0',
        'numpy==1.19.1',
    ],
    test_suite='nose.collector',
    tests_require=['nose']
)
